'''
 2019 Bjoern Annighoefer
'''

from ..value import VAL, PRM, BOL, U32, U64, STR, I64, LST, NON, EncVal
from ..query import Obj
from ..error import EOQ_ERROR_RUNTIME
#type checking
from typing import Dict, List, Any, Type, Union, Tuple
from abc import ABC

CMD_SEP = '\n'
CMD_ARG_SEP = ' '

class CMD_TYPES:
    #data related commmands
    GET = 'GET' #get cmd
    UPD = 'UPD' #update cmd
    SET = 'SET' #set cmd value #depricated 
    ADD = 'ADD' #add cmd value #depricated
    REM = 'REM' #remove cmd value #depricated
    MOV = 'MOV' #move cmd cmd #depricated
    DEL = 'DEL' #delete cmd
    CLO = 'CLO' #clone source target mode
    CRT = 'CRT' #create 
    QRF = 'QRF' #querify
    
    #comparison and merge
    CPR = 'CPR' #compare/diff
    MRG = 'MRG' #merge 
    
    #maintenance related commands
    HEL = 'HEL' #hello
    GBY = 'GBY' #goodbye
    SES = 'SES' #session
    STS = 'STS' #status
    CHG = 'CHG' #changes
    OBS = 'OBS' #observe
    UBS = 'UBS' #unobserve
    MSG = 'MSG' #send message to observers
    
    # extensions
    SCC = 'SCC' #set custom command
    CUS = 'CUS' #invoke custom command. Only used internally.
    
    # placeholders (NOT IMPLEMENTED)
    UDO = 'UDO' #undo
    RDO = 'RDO' #redo
    LCK = 'LCK' #lock
    ULK = 'ULK' #unlock

    # Verification related commands
    VER = 'VER' #verify model element against constraint 
    GAC = 'GAC' #get all applicable constraints for target
    
       
    #Result
    RES = 'RES' #result
    ERR = 'ERR' #error
    
    #Event
    EVT = 'EVT' #event
    
    #Compound commands
    CMP = 'CMP' #compound
    

class DEL_MODES:
    BAS = 'BAS' #base: delete the element itself. Does only work for elements not contained and not referenced
    AUT = 'AUT' #auto: delete the element itself. All containments and references are removed beforehand
    FUL = 'FUL' #all: delete element and all children recursively
   
    
class CLO_MODES:
    MIN = 'MIN' #minimum: clone only concept and create args
    BAS = 'BAS' #base: clone concept and all updateable all properties
    FLT = 'FLT' #removed: #flat: create the identical copy of the concept and all direct elements in the clone path
    FUL = 'FUL' #full: clone all elements in the clone path recursively

class MRG_MODES:
    STD = 'STD' #standard mode without additional information or output
    LST = 'LST' #list: creates a list with all commands
    LOG = 'LOG' #log: LST + creates a log file (info.log) with detailed information on every object
    DET = 'DET' #detailed: LOG + creates a file with all objects and their #1234 number
    
    
class EVT_TYPES:
    ELM = 'ELM' # element (this is not a real event, but used to make objects observed)
    CRT = 'CRT' # create
    UPD = 'UPD' # update
    DEL = 'DEL' # delete
    MSG = 'MSG' # message
    WAT = 'WAT' # watch event (not implemented)
    VEC = 'VEC' # verification change
    
class EVT_KEYS:
    ALL = '*' # watch all events
    
### COMMAND BASE CLASS ###    
    
class Cmd(ABC):
    def __init__(self,t:str, mute:bool=False, resName:str=None):
        self.cmd:str                = t
        self.a:List[Union[Cmd,VAL]] = []
        self.r:str                  = resName #result name
        self.m:bool                 = mute #mute the output
        
    def __eq__(self,other:Any)->bool:
        if(isinstance(other,Cmd)):
            if(self.cmd!=other.cmd or \
               self.m != other.m or \
               self.r != other.r): 
                return False
            n = len(self.a)
            m = len(other.a)
            if(n!=m):
                return False
            for i in range(n):
                if(self.a[i]!=other.a[i]):
                    return False
            return True #return true if all arguments are the same
        else:
            False 
        
    def __repr__(self)->str:
        cmdStr = ("-" if self.m else "")+str(self.cmd)
        if(len(self.a) > 0):
            cmdStr += CMD_ARG_SEP + CMD_ARG_SEP.join([str(arg) for arg in self.a])
        if(None!=self.r):
            cmdStr += " -> $"+self.r
        return cmdStr
    
    
### COMMAND REGISTRY ###

class CmdInfo:
    def __init__(self,cmdId:str, minArgs:int, maxArgs:int, description:str, clazz:Type[Cmd], unpackArgs:bool=True, optDefVals:List[Any]=[]):
        self.cmdId = cmdId
        self.minArgs = minArgs
        self.maxArgs = maxArgs
        self.description = description
        self.clazz = clazz
        self.unpackArgs = unpackArgs
        self.optArgs:int = maxArgs-minArgs
        self.optDefVals = optDefVals
        
        
EOQ_COMMAND_REGISTRY:Dict[str,CmdInfo] = {}
    

def RegisterCmdType(cmdId:str, minArgs:int, maxArgs:int, description:str, clazz:Type[Cmd], unpackArgs:bool=True, optDefVals:List[Any]=[]):
    EOQ_COMMAND_REGISTRY[cmdId] = CmdInfo(cmdId,minArgs,maxArgs,description,clazz,unpackArgs,optDefVals)
    
def CmdFactory(cmdId:str, args:List[Any], mute:bool=False, resName:str=None):
    try: 
        cmdInfo = EOQ_COMMAND_REGISTRY[cmdId]
        nArgs = len(args)
        if(nArgs < cmdInfo.minArgs or (cmdInfo.maxArgs > 0 and cmdInfo.maxArgs < nArgs)):
            raise EOQ_ERROR_RUNTIME('%s requires between %d and %d arguments, but got %d'%(cmdId,cmdInfo.minArgs,cmdInfo.maxArgs,nArgs))
        if(cmdInfo.unpackArgs):
            return cmdInfo.clazz(*args,mute=mute,resName=resName) #create a new instance
        else:
            return cmdInfo.clazz(args,mute=mute,resName=resName) #create a new instance
    except KeyError:
        return Cus(cmdId,args,mute=mute,resName=resName)
    
def GetCmdInfo(cmdId:str)->CmdInfo:
    if(cmdId in EOQ_COMMAND_REGISTRY):
        return EOQ_COMMAND_REGISTRY[cmdId]
    else:
        return None

class PrmCmd(Cmd,ABC): 
    '''Primitive command a base class for primitive commands, which fixes the argument type 
    '''
    def __init__(self,t:str, args:List[Any], mute:bool=False, resName:str=None):
        super().__init__(t, mute, resName)
        self.a:List[VAL] = [EncVal(a) for a in args]
    
### BASIC COMMANDS ###
        
class Get(PrmCmd):
    def __init__(self, target:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.GET, [target],mute,resName)     
RegisterCmdType(CMD_TYPES.GET,1,1,'Retrieve values',Get,True)
        
class Upd(PrmCmd):
    def __init__(self, target:Any, feature:Any, value:Any, position:Any=0,mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.UPD, [target,feature,value,I64(position)],mute,resName)
RegisterCmdType(CMD_TYPES.UPD,3,4,'Change, add or delete the value of a feature.',Upd,True,[-10]) #-10 = hack to prevent argument stripping 
        
class Del(PrmCmd):
    def __init__(self, target:Any, mode:Any=DEL_MODES.BAS,mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.DEL, [target,mode],mute,resName)
RegisterCmdType(CMD_TYPES.DEL,1,2,'Delete an element.',Del,True,[DEL_MODES.BAS]) 
        
class Set(PrmCmd):
    def __init__(self, target:Any, feature:Any, value:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.SET, [target,feature,value],mute,resName)
RegisterCmdType(CMD_TYPES.SET,3,3,'Set a feature.',Set,True) 
        
class Add(PrmCmd):
    def __init__(self, target:Any, feature:Any, value:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.ADD, [target,feature,value],mute,resName)
RegisterCmdType(CMD_TYPES.ADD,3,3,'Add values to a multi-value feature.',Add,True) 
        
class Rem(PrmCmd):
    def __init__(self, target:Any, feature:Any, value:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.REM, [target,feature,value],mute,resName)
RegisterCmdType(CMD_TYPES.REM,3,3,'Remove a value from a feature.',Rem,True) 
        
class Mov(PrmCmd):
    def __init__(self, target:Any, newIndex:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.MOV, [target,newIndex],mute,resName)
RegisterCmdType(CMD_TYPES.MOV,2,2,'Remove a value from a feature.',Mov,True) 
        
class Clo(PrmCmd):
    def __init__(self, target:Any, mode:Any=STR(CLO_MODES.MIN), createArgOverrides:Any=LST([]), mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.CLO,[target,mode,createArgOverrides],mute,resName)
RegisterCmdType(CMD_TYPES.CLO,1,3,'Clone an element.',Clo,True,[CLO_MODES.MIN,[]]) 
          
class Crt(PrmCmd):
    def __init__(self, classId:Any, n:Any=U32(1), createArgs:Any=LST([]), target:Any=NON(), recoveryArgs:Any=LST([]),mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.CRT,[classId,n,createArgs,target,recoveryArgs],mute,resName)
RegisterCmdType(CMD_TYPES.CRT,2,5,'Create element by class ID (name or Obj).',Crt,True,[[],None,[]]) 
        
class Cpr(PrmCmd):
    def __init__(self, dst:Any, src:Any, mode:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.CPR,[dst,src,mode],mute,resName)
RegisterCmdType(CMD_TYPES.CPR,3,3,'Compare two element and retrieve the difference.',Cpr,True) 
        
class Mrg(PrmCmd):
    def __init__(self, dst:Any, src:Any, mode:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.MRG,[dst,src,mode],mute,resName)
RegisterCmdType(CMD_TYPES.MRG,3,3,'Merge all changes of new into old.',Mrg,True) 
        
### DOMAIN RELATED COMMANDS
        
class Hel(PrmCmd):
    def __init__(self, user:Any, password:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.HEL, [user, password],mute,resName)
RegisterCmdType(CMD_TYPES.HEL,2,2,'Log into the domain. The login persists for the actual session id.',Hel,True) 
        
class Ses(PrmCmd):
    def __init__(self, sessionId:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.SES, [sessionId],mute,resName)
RegisterCmdType(CMD_TYPES.SES,1,1,'Set the session id for the current transaction. Only makes sense in a compound command.',Ses,True) 
        
class Gby(PrmCmd):
    def __init__(self, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.GBY, [],mute,resName)
RegisterCmdType(CMD_TYPES.GBY,0,0,'Log off the given session.',Gby,True) 
        
class Sts(PrmCmd):
    def __init__(self, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.STS, [],mute,resName)
RegisterCmdType(CMD_TYPES.STS,0,0,'Status returns the latest change ID.',Sts,True) 
        
class Chg(PrmCmd):
    def __init__(self, latestChangeId:Any, n:Any, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.CHG,[latestChangeId,n],mute,resName)
RegisterCmdType(CMD_TYPES.CHG,2,2,'Retrieve change records.',Chg,True) 
        
class Obs(PrmCmd):
    def __init__(self, eventType:Any, eventKey:Any=None, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.OBS, [eventType,eventKey],mute,resName)
RegisterCmdType(CMD_TYPES.OBS,1,2,'Start observing events for the current session.',Obs,True,[None]) 
        
class Ubs(PrmCmd):
    def __init__(self, eventType:Any, eventKey:Any=None, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.UBS, [eventType,eventKey],mute,resName)
RegisterCmdType(CMD_TYPES.UBS,1,2,'Stop observing events for the current session.',Ubs,True,[None]) 

class Msg(PrmCmd):
    def __init__(self, msgKey:Any, msg:Any, mute:bool=True,resName : str = None):
        super().__init__(CMD_TYPES.MSG, [msgKey,msg],mute,resName)
RegisterCmdType(CMD_TYPES.MSG,2,2,'Sending a message to observers.',Msg,True) 
        
# class Lck(Cmd):
#     def __init__(self,timeout : float,mute:bool=False, resName:str=None):
#         super().__init__(CMD_TYPES.LCK, [timeout],mute,resName)
# RegisterCmdType(CMD_TYPES.LCK,1,1,'Lock the domain to the active session.',Lck,True) 
#         
# class Ulk(Cmd):
#     def __init__(self,mute:bool=False, resName:str=None):
#         super().__init__(CMD_TYPES.LCK, [],mute,resName)
# RegisterCmdType(CMD_TYPES.ULK,0,0,'Unlock release the domain lock.',Ulk,True) 
        
        
### CONSTRAINT RELATED COMMANDS ###

class Ver(PrmCmd):
    def __init__(self, target:Any, mute: bool=False,resName : str = None):
        super().__init__(CMD_TYPES.VER, [target],mute,resName)
RegisterCmdType(CMD_TYPES.VER,1,1,'Verify',Ver,True)

class Gac(PrmCmd):
    def __init__(self, target:Any, mute: bool=False,resName : str = None):
        super().__init__(CMD_TYPES.GAC, [target],mute,resName)
RegisterCmdType(CMD_TYPES.GAC,1,1,'Get all constraints',Gac,True)
        
        
### RESULT COMMAND ###

class Res(PrmCmd):
    def __init__(self, tid:Any, cid:Any, val:Any, resNames:List[str]=[],resNameIndizies:List[int]=[],mute: bool=False,resName : str = None):
        super().__init__(CMD_TYPES.RES, [U64(tid),U64(cid),val,resNames,resNameIndizies],mute,resName)
RegisterCmdType(CMD_TYPES.RES,3,5,'Result',Res,True,[[],[]]) 
        
class Err(PrmCmd):
    def __init__(self, code:int,msg:Any='', debug:Any='',mute: bool=False, resName:str=None):
        super().__init__(CMD_TYPES.ERR, [U32(code),STR(msg),STR(debug)],mute,resName)
RegisterCmdType(CMD_TYPES.ERR,1,3,'Error result',Err,True,['','']) 
        
        
### EVENT COMMANDS ###

class Evt(PrmCmd):
    def __init__(self, eventType:STR, data:LST, mute:bool=False, resName:str=None):
        super().__init__(CMD_TYPES.EVT, [eventType, data])
RegisterCmdType(CMD_TYPES.EVT,2,2,'Event',Evt,True)
        
class UpdEvt(Evt):
    def __init__(self, cid:U64, target:Obj, feature:STR, value:PRM, position:I64, user:STR):
        super().__init__(STR(EVT_TYPES.UPD),LST([cid,target,feature,value,position,user]))
        
class CrtEvt(Evt):
    def __init__(self, cid:U64, target:Obj, classId:Tuple[STR,Obj], createArgs:LST, user:STR):
        super().__init__(STR(EVT_TYPES.CRT),LST([cid,target,classId,createArgs,user]))
        
class DelEvt(Evt):
    def __init__(self, cid:U64, target:Obj, user:STR):
        super().__init__(STR(EVT_TYPES.DEL),LST([cid,target,user]))
        
class MsgEvt(Evt):
    def __init__(self, msgKey:STR, msg:STR):
        super().__init__(STR(EVT_TYPES.MSG),LST([msgKey, msg]))
        
class WatEvt(Evt):
    def __init__(self, wid:U32, result:VAL, error:STR):
        super().__init__(STR(EVT_TYPES.WAT),LST([wid, result, error]))
        
class VecEvt(Evt):
    def __init__(self, target:Obj,  constraint:Obj, isValid:BOL,  error:STR):
        super().__init__(STR(EVT_TYPES.VEC),LST([target,constraint,isValid,error]))
        
        
### CUSTOM COMMAND ###

class Scc(PrmCmd):
    def __init__(self, cmdId:Any, cmdStr:Any, mute:bool=False, resName:str = None):
        super().__init__(CMD_TYPES.SCC, [cmdId,cmdStr],mute,resName)
RegisterCmdType(CMD_TYPES.SCC,2,2,'Set custom command',Scc,True) 


class Cus(PrmCmd):
    def __init__(self, cmdId:Any, args:List[any], mute:bool=False, resName:str=None):
        super().__init__(cmdId, args, mute, resName)
#RegisterCmdType(CMD_TYPES.CUS,1,1,'Custom command',Cus,True) 
    
        

### COMPOUND COMMAND ###

class Cmp(Cmd):
    def __init__(self, cmds:List[Cmd]=None ,mute: bool=False, resName:str=None):
        super().__init__(CMD_TYPES.CMP,mute,resName)
        self.a:Cmd = []
        if(None != cmds):
            self.a = [c for c in cmds if isinstance(c,Cmd)]
        
    def Get(self, target:Any, mute:bool=False, resName:str=None):
        self.a.append(Get(target,mute,resName))
        return self
    
    def Upd(self, target:Any, feature:Any, value:Any, position:Any=0, mute:bool=False, resName:str=None):
        self.a.append(Upd(target,feature,value,position,mute,resName))
        return self
    
    def Del(self, target:Any, mode:Any=DEL_MODES.BAS, mute:bool=False, resName:str=None):
        self.a.append(Del(target,mode,mute,resName))
        return self
    
    def Set(self, target:Any, feature:Any, value:Any, mute:bool=False, resName:str=None):
        self.a.append(Set(target,feature,value,mute,resName))
        return self
    
    def Add(self, target:Any, feature:Any, value:Any, mute:bool=False, resName:str=None):
        self.a.append(Add(target,feature,value,mute,resName))
        return self
    
    def Rem(self, target:Any, feature:Any, value:Any, mute:bool=False, resName:str=None):
        self.a.append(Rem(target,feature,value,mute,resName))
        return self
    
    def Mov(self, target:Any, newIndex:Any, mute:bool=False, resName:str=None):
        self.a.append(Mov(target,newIndex,mute,resName))
        return self
    
    def Clo(self, target:Any, mode:Any=STR(CLO_MODES.MIN), createArgOverrides:Any=LST([]), mute:bool=False, resName:str=None):
        self.a.append(Clo(target,mode,createArgOverrides,mute,resName))
        return self
    
    def Crt(self, classId:Any, n:Any=U32(1), createArgs:Any=LST([]), target:Any=NON(), recoveryArgs:Any=LST([]), mute:bool=False, resName:str=None):
        self.a.append(Crt(classId,n,createArgs,target,recoveryArgs,mute,resName))
        return self
    
    def Mrg(self, dst:Any, src:Any, mode:Any, mute:bool=False, resName:str=None):
        self.a.append(Mrg(dst,src,mode,mute,resName))
        return self

    def Cpr(self,dst,src,mode,mute:bool=False, resName:str=None):
        self.a.append(Cpr(dst,src,mode,mute,resName))
        return self
    
    def Hel(self, user:Any, password:Any, mute:bool=False, resName:str=None):
        self.a.append(Hel(user,password,mute,resName))
        return self
    
    def Ses(self, sessionId:Any, mute:bool=False, resName:str=None):
        self.a.append(Ses(sessionId,mute,resName))
        return self
    
    def Gby(self, mute:bool=False, resName:str=None):
        self.a.append(Gby(mute,resName))
        return self
    
    def Sts(self, mute:bool=False, resName:str=None):
        self.a.append(Sts(mute,resName))
        return self
    
    def Chg(self, changeId:Any, n:Any, mute:bool=False, resName:str=None):
        self.a.append(Chg(changeId,n,mute,resName))
        return self
    
    def Obs(self, eventType:Any, eventKey:Any=None, mute:bool=False, resName:str=None):
        self.a.append(Obs(eventType,eventKey,mute,resName))
        return self
    
    def Ubs(self, eventType:Any, eventKey:Any=None, mute:bool=False, resName:str=None):
        self.a.append(Ubs(eventType,eventKey,mute,resName))
        return self

    def Ver(self, target:Any):
        self.a.append(Ver(target))
        return self

    def Gac(self,target:Any):
        self.a.append(Gac(target))
        return self
    
    def Scc(self, cmdId:Any, cmdStr:Any, mute:bool=False, resName:str = None):
        self.a.append(Scc(cmdId,cmdStr,mute,resName))
        return self
    
    def Cus(self, cmdId:Any, args:List[Any], mute:bool=False, resName:str=None):
        self.a.append(Cus(cmdId,args,mute=mute,resName=resName))
        return self
        
    def Append(self, cmd:Cmd):
        '''Appends a command to the compound
        '''
        self.a.append(cmd)
        return self
            
RegisterCmdType(CMD_TYPES.CMP,0,-1,'A compound command composed of a list of subcommands.',Cmp,False) 

### HELPERS ###

def IsReadOnlyCmd(cmd:Cmd)->bool:
    ''' Returns true for commands that do not modify the MDB
    '''
    if(isinstance(cmd,Get)): #Get is the only option for non changing commands
        return True
    elif(isinstance(cmd,Cmp)): #in case of compound commands it depends on the subcommands
        for a in cmd.a:
            if(not IsReadOnlyCmd(a)):
                return False
        return True
    else:
        return False #all other commands are expected to have changes
    






'''
Definition of queries and segments 
 
Bjoern Annighoefer 2024
'''

from ..value import VAL, STR, I64, QRY, EncVal, VAL_TYPES, RegisterValType
from ..error import EOQ_ERROR_INVALID_TYPE
#type checking
import typing


class SEG_TYPES:
    OBJ = 'OBJ' #object reference
    ONI = 'ONI' #object by name or id
    HIS = 'HIS' #history reference
    
    # Left only element-wise operations
    PTH = 'PTH' # path
    CLS = 'CLS' # class
    INO = 'INO' # instance of
    MET = 'MET' # meta
    NOT = 'NOT' # boolean not
    TRM = 'TRM' # terminate
    TRY = 'TRY' #try catch
    
    QRF = 'QRF' #querify
    STF = 'STF' #stringify
    SLF = 'SLF' #self: returns the context without modifications

    # Left only list-of-elements-wise operations
    IDX = 'IDX' # index
    SEL = 'SEL' # selector
    ARR = 'ARR' # outer array
    ZIP = 'ZIP' # inner array creation
    ANY = 'ANY' # at least one element from the right is found in the left --> Bool
    ALL = 'ALL' # all elements from the right are found in the left --> Bool
    
    # Left-vs-right element-wise operators
    EQU = 'EQU' # equal
    EQA = 'EQA' # equal any of the right elements
    NEQ = 'NEQ' # not equal
    LES = 'LES' # less
    GRE = 'GRE' # greater
    RGX = 'RGX' # regex (string only)
    
    ##generic operators
    ADD = 'ADD' # OR, addition,
    SUB = 'SUB' # XOR, subtraction,
    MUL = 'MUL' # AND, multiply,
    DIV = 'DIV' # NAND, divided,
    
    ##logic operator synonyms
    ORR = 'ORR' # OR synonym
    XOR = 'XOR' # XOR synonym
    AND = 'AND' # AND synonym
    NAD = 'NAD' # NAND synonym
    
    #Left-vs-right list-of-element-wise operators
    CSP = 'CSP' # cross product
    ITS = 'ITS' # intersection
    DIF = 'DIF' # set subtraction / relative complement
    UNI = 'UNI' # union
    CON = 'CON' # concat
        

# META OPERATORS 
# Use only three letters for meta operators

#element operators
class MET_MODES:
    CLS = 'CLASS' #class
    CLN = 'CLASSNAME' #class name
    CON = 'CONTAINER' #parent (container)
    PAR = 'PARENT' #parent (container)
    ALP = 'ALLPARENTS' #parent (container)
    ASO = 'ASSOCIATES' #ASSOCIATES(start=root) all elements refering to this one beginning at start. default is root
    IDX = 'INDEX' #index within its containment
    CFT = 'CONTAININGFEATURE' #the feature that contains the element
    FEA = 'FEATURES' #all features
    FEV = 'FEATUREVALUES' #all feature values
    FEN = 'FEATURENAMES' #all feature names
    ATT = 'ATTRIBUTES' #all attribute features
    ATN = 'ATTRIBUTENAMES' #all attribute feature names
    ATV = 'ATTRIBUTEVALUES' #all attribute feature values
    REF = 'REFERENCES' #all reference features
    REN = 'REFERENCENAMES' #all reference feature names
    REV = 'REFERENCEVALUES' #all reference feature values
    CNT = 'CONTAINMENTS' #all containment features
    CNV = 'CONTAINMENTVALUES' #all containment feature values
    CNN = 'CONTAINMENTNAMES' #all containment feature names
    
    #class operators
    PAC = 'PACKAGE' #class
    STY = 'SUPERTYPES' #directly inherited classes
    ALS = 'ALLSUPERTYPES' #all and also indirectly inherited classes
    IMP = 'IMPLEMENTERS' #all direct implementers of a class
    ALI = 'ALLIMPLEMENTERS' #all and also indirect implementers of a class  
    MMO = 'METAMODELS' #retrieve all metamodels
    
    #Control flow operators 
    IFF = 'IF' #if(condition,then,else);  #DEPRICATED
    TRY = 'TRY' #catch errors and return a default #NOT IMPLEMENTED
    
    
    #list operators
    LEN = 'SIZE' #size of a list #DEPRICATED
    
    #recursive operators
    REC = 'REPEAT' #REPEAT(<query>,depth) repeat a given query until no more results are found #NOT IMPLEMENTED
    
    
    
class IDX_MODES:
    #structure operators
    FLT = 'FLATTEN' #flatten any sub list structure to a list #NOT IMPLEMENTED
    LEN = 'SIZE' #size of a list
    ASC = 'SORTASC' #sort ascending #NOT IMPLEMENTED
    DSC = 'SORTDSC' #sort descending #NOT IMPLEMENTED
    
    
class STF_MODES:
    FUL = 'FUL' #full, only a single string is returned
    LST = 'LST' #only the elements of a list are stringified on the uppermost level
    ELM = 'ELM' #elementwise

        
class Seg:
    def __init__(self,stype,args):
        self.seg = stype
        self.v = [EncVal(v) for v in args]
        
    def __repr__(self):
        s = '&' + self.seg
        nArgs = len(self.v)
        if(nArgs > 0):
            s += str(self.v[0])
        for i in range(1,nArgs):
            s += ';'+str(self.v[i])
        return s
        
    def __eq__(self, other):
        if(isinstance(other, Seg)):
            return self.seg == other.seg and self.v == other.v
        else: 
            return False #can not be equal if it is a different type
        
    def __lt__(self, other):
        if(isinstance(other,Seg)):
            #first compare the segment identifier
            if(self.seg != other.seg):
                return self.seg < other.seg
            #second compare the values
            n = len(self.v)
            m = len(other.v)
            #compare segments sequentially for equality
            for i in range(min(n,m)):
                if(self.v[i] != other.v[i]):
                    return self.v[i] < other.v[i]
            return n < m
        else:
            raise EOQ_ERROR_INVALID_TYPE("< not supported between instances of 'Seg' and '%d'."%(type(other).__name__))
        
    def __gt__(self, other):
        if(isinstance(other,Seg)):
            #first compare the segment identifier
            if(self.seg != other.seg):
                return self.seg > other.seg
            #second compare the values
            n = len(self.v)
            m = len(other.v)
            #compare segments sequentially for equality
            for i in range(min(n,m)):
                if(self.v[i] != other.v[i]):
                    return self.v[i] > other.v[i]
            return n > m
        else:
            raise EOQ_ERROR_INVALID_TYPE("< not supported between instances of 'Seg' and '%d'."%(type(other).__name__))
        
    def __hash__(self):
        return hash(self.__repr__())
    
    
        
class Qry(QRY): #Essential object cmd 
    def __init__(self,value=None):
        QRY.__init__(self)
        self.v:typing.List[Seg] = []
        if(None == value):
            self.v = [] #empty list of segments
        elif(isinstance(value, Seg)):
            self.v = [value]
        elif(isinstance(value, list) and all([isinstance(s,Seg) for s in value])):
            self.v = [s for s in value]
        elif(isinstance(value, list) and all([isinstance(s,Qry) for s in value])):
            self.Arr([s for s in value])
        elif(isinstance(value, Qry)): #Clone operator
            self.v = [s for s in value.v]
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected None, Seg, list<Seg>, list<Obj> or Qry but got %s"%(value))
        
    def __repr__(self):
        queryStr = ''
        for seg in self.v:
            queryStr += str(seg)
        return '('+queryStr+')'

    def __lt__(self, other):
        if(isinstance(other,QRY)):
            n = len(self.v)
            m = len(other.v)
            #compare segments sequentially for equality
            for i in range(min(n,m)):
                if(self.v[i] != other.v[i]):
                    return self.v[i] < other.v[i]
            return n < m
        else:
            raise EOQ_ERROR_INVALID_TYPE("< not supported between instances of 'Qry' and '%d'."%(type(other).__name__))
    
    def __gt__(self, other):
        if(isinstance(other,QRY)):
            n = len(self.v)
            m = len(other.v)
            #compare segments sequentially for equality
            for i in range(min(n,m)):
                if(self.v[i] != other.v[i]):
                    return self.v[i] > other.v[i]
            return n > m
        else:
            raise EOQ_ERROR_INVALID_TYPE("> not supported between instances of 'Qry' and '%d'."%(type(other).__name__))
    
    def _(self,seg): #adds an existing segment
        self.v.append(seg)
    
    def Obj(self,v):
        if(isinstance(v,Seg)):
            self.v.append(v)
        else:
            self.v.append(Seg(SEG_TYPES.OBJ,[v]))
        return self
    
    def Oni(self,v):
        self.v.append(Seg(SEG_TYPES.ONI,[v]))
        return self
    
    def His(self,v):
        self.v.append(Seg(SEG_TYPES.HIS,[v]))
        return self
        
    def Pth(self,name):
        self.v.append(Seg(SEG_TYPES.PTH,[name]))
        return self
    
    def Cls(self,clazz):
        self.v.append(Seg(SEG_TYPES.CLS,[clazz]))
        return self
    
    def Ino(self,clazz):
        self.v.append(Seg(SEG_TYPES.INO,[clazz]))
        return self
    
    def Met(self,name,args=[]):
        self.v.append(Seg(SEG_TYPES.MET,[name, args]))
        return self
    
    def Not(self):
        self.v.append(Seg(SEG_TYPES.NOT,[]))
        return self
    
    def Trm(self,cond=None,fallback=None):
        self.v.append(Seg(SEG_TYPES.TRM,[cond,fallback]))
        return self
    
    def Qrf(self):
        self.v.append(Seg(SEG_TYPES.QRF,[]))
        return self
    
    def Slf(self):
        self.v.append(Seg(SEG_TYPES.SLF,[]))
        return self
    
    def Stf(self, mode=STF_MODES.ELM):
        self.v.append(Seg(SEG_TYPES.STF,[mode]))
        return self
    
    def Try(self,query,fallback=None):
        self.v.append(Seg(SEG_TYPES.TRY,[query,fallback]))
        return self
    
    def Idx(self,start,stop=None,step=None):
        if(None == stop):
            self.v.append(Seg(SEG_TYPES.IDX,[start]))
        elif(None == step):
            self.v.append(Seg(SEG_TYPES.IDX,[start,stop]))
        else:
            self.v.append(Seg(SEG_TYPES.IDX,[start,stop,step]))
        return self
    
    def Sel(self,query):
        self.v.append(Seg(SEG_TYPES.SEL,[query]))
        return self
    
    def Arr(self,elements):
        self.v.append(Seg(SEG_TYPES.ARR,[elements]))
        return self

    def Zip(self,elements):
        self.v.append(Seg(SEG_TYPES.ZIP,[elements]))
        return self
    
    def Any(self,query):
        self.v.append(Seg(SEG_TYPES.ANY,[query]))
        return self
    
    def All(self,query):
        self.v.append(Seg(SEG_TYPES.ALL,[query]))
        return self
    
    def Equ(self,query):
        self.v.append(Seg(SEG_TYPES.EQU,[query]))
        return self
    
    def Eqa(self,query):
        self.v.append(Seg(SEG_TYPES.EQA,[query]))
        return self
    
    def Neq(self,query):
        self.v.append(Seg(SEG_TYPES.NEQ,[query]))
        return self
    
    def Les(self,query):
        self.v.append(Seg(SEG_TYPES.LES,[query]))
        return self
    
    def Gre(self,query):
        self.v.append(Seg(SEG_TYPES.GRE,[query]))
        return self
        
    def Rgx(self,query):
        self.v.append(Seg(SEG_TYPES.RGX,[query]))
        return self
    
    def Add(self,query):
        self.v.append(Seg(SEG_TYPES.ADD,[query]))
        return self
    
    def Sub(self,query):
        self.v.append(Seg(SEG_TYPES.SUB,[query]))
        return self
    
    def Mul(self,query):
        self.v.append(Seg(SEG_TYPES.MUL,[query]))
        return self
    
    def Div(self,query):
        self.v.append(Seg(SEG_TYPES.DIV,[query]))
        return self
    
    def Orr(self,query):
        self.v.append(Seg(SEG_TYPES.ORR,[query]))
        return self
    
    def Xor(self,query):
        self.v.append(Seg(SEG_TYPES.XOR,[query]))
        return self
    
    def And(self,query):
        self.v.append(Seg(SEG_TYPES.AND,[query]))
        return self
    
    def Nad(self,query):
        self.v.append(Seg(SEG_TYPES.NAD,[query]))
        return self
    
    def Csp(self,query):
        self.v.append(Seg(SEG_TYPES.CSP,[query]))
        return self
    
    def Its(self,query):
        self.v.append(Seg(SEG_TYPES.ITS,[query]))
        return self
    
    def Dif(self,query):
        self.v.append(Seg(SEG_TYPES.DIF,[query]))
        return self
    
    def Uni(self,query):
        self.v.append(Seg(SEG_TYPES.UNI,[query]))
        return self
    
    def Con(self,query):
        self.v.append(Seg(SEG_TYPES.CON,[query]))
        return self
RegisterValType(VAL_TYPES.QRY,'EOQ query',Qry)
    
''' Shortcuts to start queries '''
    
class Obj(Qry):
    def __init__(self,v):
        super().__init__()
        self.Obj(v)
        
class Oni(Qry):
    def __init__(self,v):
        super().__init__()
        self.Oni(v)
        
class His(Qry):
    def __init__(self,v):
        super().__init__()
        self.His(v)

class Pth(Qry):
    def __init__(self,name):
        super().__init__()
        self.Pth(name)
    
class Cls(Qry):
    def __init__(self,clazz):
        super().__init__()
        self.Cls(clazz)
        
class Ino(Qry):
    def __init__(self,clazz):
        super().__init__()
        self.Ino(clazz)
        
class Try(Qry):
    def __init__(self,query,fallback):
        super().__init__()
        self.Try(query,fallback)
        
class Met(Qry):
    def __init__(self,name,args=[]):
        super().__init__()
        self.Met(name,args)
        
class Qrf(Qry):
    def __init__(self):
        super().__init__()
        self.Qrf()
        
class Stf(Qry):
    def __init__(self, mode=STF_MODES.ELM):
        super().__init__()
        self.Stf(mode)
        
class Slf(Qry):
    def __init__(self):
        super().__init__()
        self.Slf()
        
class Not(Qry):
    def __init__(self):
        super().__init__()
        self.Not()
        
class Idx(Qry):
    def __init__(self,start,stop=None,step=None):
        super().__init__()
        self.Idx(start,stop,step)
        
class Sel(Qry):
    def __init__(self,query):
        super().__init__()
        self.Sel(query)
        
class Arr(Qry):
    def __init__(self,elements):
        super().__init__()
        self.Arr(elements)       
    
class Zip(Qry):
    def __init__(self,elements):
        super().__init__()
        self.Zip(elements)   
        
class Any(Qry):
    def __init__(self,select):
        super().__init__()
        self.Any(select)
        
class All(Qry):
    def __init__(self,select):
        super().__init__()
        self.All(select)
        
class Equ(Qry):
    def __init__(self,query):
        super().__init__()
        self.Equ(query)
        
class Eqa(Qry):
    def __init__(self,query):
        super().__init__()
        self.Eqa(query)
        
class Neq(Qry):
    def __init__(self,query):
        super().__init__()
        self.Neq(query)
        
class Les(Qry):
    def __init__(self,query):
        super().__init__()
        self.Les(query)
        
class Gre(Qry):
    def __init__(self,query):
        super().__init__()
        self.Gre(query)
        
class Rgx(Qry):
    def __init__(self,query):
        super().__init__()
        self.Rgx(query)
        
        
### HELPERS ###
def IsObj(e:typing.Any)->bool:
    '''Checks if an Query is a special Object query
    '''
    return (isinstance(e,QRY) and 1==len(e.v) and SEG_TYPES.OBJ == e.v[0].seg)

def ObjId(e:typing.Any)->VAL:
    '''Returns safely the framework-specific object ID
    '''
    if(IsObj(e)):
        return e.v[0].v[0]
    else:
        raise EOQ_ERROR_INVALID_TYPE("Requires element of type Obj.")
    
def IsHis(e:typing.Any)->bool:
    '''Checks if an Query is a special Object query
    '''
    return (isinstance(e,QRY) and 1==len(e.v) and SEG_TYPES.HIS == e.v[0].seg)

def HisId(e:typing.Any)->typing.Tuple[STR,I64]:
    '''Returns safely history Id
    '''
    if(IsHis(e)):
        return e.v[0].v[0]
    else:
        raise EOQ_ERROR_INVALID_TYPE("Requires element of type His.")



'''
 Bjoern Annighoefer 2019
'''

from .serializer import Serializer, RegisterSerializer
'''

JSON Serializer

'''
   
from ..frame import Frm
from ..command import Cmd, CmdFactory
from ..query import Qry, Seg
from ..value import VAL, ValFactory, U64, I64, VAL_TYPES
from ..error import EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_RUNTIME

import json
#from unittest.mock import patch

## JSON float formating patch: https://stackoverflow.com/questions/1447287/format-floats-with-standard-json-module
# We need to ensure that c encoder will not be launched
# @patch('json.encoder.c_make_encoder', None)
# def JsonDumpsWithWithControlledFloatFormat(obj, **kw):
#     # saving original method
#     of = json.encoder._make_iterencode
#     def inner(*args, **kwargs):
#         args = list(args)
#         # fifth argument is float formater which will we replace
#         args[4] = lambda o: ('%f'%(o)).rstrip('0').rstrip('.') #remove trailing 0 and .
#         return of(*args, **kwargs)
#     
#     with patch('json.encoder._make_iterencode', wraps=inner):
#         return json.dumps(obj, **kw)

def JsonSerializeHook(obj):
    if(isinstance(obj,Frm)):
        return {"frm":"ami","ver":obj.ver,"typ":obj.typ,"uid":obj.uid,"sid":obj.sid,"ser":obj.ser,"roc":obj.roc,"dat":obj.dat}
    elif(isinstance(obj,Cmd)):
        dct = {"cmd":obj.cmd,"a":obj.a}
        #serialize mute and name only if given
        if(obj.m): dct["m"] = obj.m
        if(obj.r): dct["r"] = obj.r
        return dct
    elif(isinstance(obj,VAL)):
        if(type(obj) in (U64,I64)): #64 integer need special care, because JSON does not support BigInt
            return {"val":obj.val,"v":str(obj.v)}
        else:
            return {"val":obj.val,"v":obj.v}
    elif(isinstance(obj,Seg)):
        return {"seg":obj.seg,"v":obj.v}
    else:
        raise EOQ_ERROR_INVALID_TYPE('Can not serialize element of type: %s'%(type(obj).__name__))

def JsonDeserializeHook(dct):
        if "frm" in dct:
            return Frm(dct["typ"],int(dct["uid"]),dct["ser"],dct["dat"],dct["sid"],dct["roc"],dct["ver"])
        if "cmd" in dct:
            cmdType = dct["cmd"]
            a = dct["a"]
            m = False
            r = None
            if "m" in dct: m = dct["m"]
            if "r" in dct: r = dct["r"]
            cmd = CmdFactory(cmdType, a, m, r)
            return cmd
        elif "val" in dct: 
            valType = dct["val"]
            value = None
            if(valType in [VAL_TYPES.U64, VAL_TYPES.I64]): #64 integer need special care, because JSON does not support BigInt
                value = int(dct["v"])
            else:
                value = dct["v"]
            val = ValFactory(valType,value)
            return val
        elif "seg" in dct: 
            return Seg(dct["seg"],dct["v"])
        return dct


class JsonSerializer(Serializer):
    def Name(self):
        '''Returns a three letter identifier for this kind of serialization 
        '''
        return 'JSO'
    
    #@override
    def Ser(self, val):
        try:
            return json.dumps(val,separators=(',', ':'),default=JsonSerializeHook,ensure_ascii=False)
            #return JsonDumpsWithWithControlledFloatFormat(val,separators=(',', ':'),default=JsonSerializeHook) #disabled, because it slows down json dumps by a factor of 5
        except Exception as e:
            raise EOQ_ERROR_RUNTIME('Serialization failed: %s'%(str(e)))
        
    #@override    
    def SerFrm(self,frm : Frm):
        return self.Ser(frm)
    
    #@override    
    def SerCmd(self, cmd : Cmd):
        return self.Ser(cmd)
    
    #@override
    def SerQry(self, qry : Qry):
        return self.Ser(qry)
    
    #@override
    def SerSeg(self, seg : Seg):
        return self.Ser(seg)
    
    #@override
    def SerVal(self,val):
        return self.Ser(val)
    
    #@override
    def Des(self,code):
        return json.loads(code,object_hook=JsonDeserializeHook)
    
    #@override
    def DesFrm(self,data) -> Frm:
        return self.Des(data)
        
    #@override
    def DesCmd(self,data) -> Cmd:
        return self.Des(data)
    
    #@override
    def DesQry(self,data) -> Qry:
        return self.Des(data)
    
    #@override
    def DesSeg(self,data) -> Seg:
        return self.Des(data)
    
    #@override
    def DesVal(self,data):
        return self.Des(data)

### make this serializer known globally 
RegisterSerializer(JsonSerializer)
    



# Generated from ../eoq3.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .eoq3Parser import eoq3Parser
else:
    from eoq3Parser import eoq3Parser

# This class defines a complete listener for a parse tree produced by eoq3Parser.
class eoq3Listener(ParseTreeListener):

    # Enter a parse tree produced by eoq3Parser#cmds.
    def enterCmds(self, ctx:eoq3Parser.CmdsContext):
        pass

    # Exit a parse tree produced by eoq3Parser#cmds.
    def exitCmds(self, ctx:eoq3Parser.CmdsContext):
        pass


    # Enter a parse tree produced by eoq3Parser#cmd.
    def enterCmd(self, ctx:eoq3Parser.CmdContext):
        pass

    # Exit a parse tree produced by eoq3Parser#cmd.
    def exitCmd(self, ctx:eoq3Parser.CmdContext):
        pass


    # Enter a parse tree produced by eoq3Parser#cmd_arg.
    def enterCmd_arg(self, ctx:eoq3Parser.Cmd_argContext):
        pass

    # Exit a parse tree produced by eoq3Parser#cmd_arg.
    def exitCmd_arg(self, ctx:eoq3Parser.Cmd_argContext):
        pass


    # Enter a parse tree produced by eoq3Parser#cmd_res_name.
    def enterCmd_res_name(self, ctx:eoq3Parser.Cmd_res_nameContext):
        pass

    # Exit a parse tree produced by eoq3Parser#cmd_res_name.
    def exitCmd_res_name(self, ctx:eoq3Parser.Cmd_res_nameContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val.
    def enterVal(self, ctx:eoq3Parser.ValContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val.
    def exitVal(self, ctx:eoq3Parser.ValContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_prim.
    def enterVal_prim(self, ctx:eoq3Parser.Val_primContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_prim.
    def exitVal_prim(self, ctx:eoq3Parser.Val_primContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_bol.
    def enterVal_bol(self, ctx:eoq3Parser.Val_bolContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_bol.
    def exitVal_bol(self, ctx:eoq3Parser.Val_bolContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_u32.
    def enterVal_u32(self, ctx:eoq3Parser.Val_u32Context):
        pass

    # Exit a parse tree produced by eoq3Parser#val_u32.
    def exitVal_u32(self, ctx:eoq3Parser.Val_u32Context):
        pass


    # Enter a parse tree produced by eoq3Parser#val_u64.
    def enterVal_u64(self, ctx:eoq3Parser.Val_u64Context):
        pass

    # Exit a parse tree produced by eoq3Parser#val_u64.
    def exitVal_u64(self, ctx:eoq3Parser.Val_u64Context):
        pass


    # Enter a parse tree produced by eoq3Parser#val_i32.
    def enterVal_i32(self, ctx:eoq3Parser.Val_i32Context):
        pass

    # Exit a parse tree produced by eoq3Parser#val_i32.
    def exitVal_i32(self, ctx:eoq3Parser.Val_i32Context):
        pass


    # Enter a parse tree produced by eoq3Parser#val_i64.
    def enterVal_i64(self, ctx:eoq3Parser.Val_i64Context):
        pass

    # Exit a parse tree produced by eoq3Parser#val_i64.
    def exitVal_i64(self, ctx:eoq3Parser.Val_i64Context):
        pass


    # Enter a parse tree produced by eoq3Parser#val_f32.
    def enterVal_f32(self, ctx:eoq3Parser.Val_f32Context):
        pass

    # Exit a parse tree produced by eoq3Parser#val_f32.
    def exitVal_f32(self, ctx:eoq3Parser.Val_f32Context):
        pass


    # Enter a parse tree produced by eoq3Parser#val_f64.
    def enterVal_f64(self, ctx:eoq3Parser.Val_f64Context):
        pass

    # Exit a parse tree produced by eoq3Parser#val_f64.
    def exitVal_f64(self, ctx:eoq3Parser.Val_f64Context):
        pass


    # Enter a parse tree produced by eoq3Parser#val_str.
    def enterVal_str(self, ctx:eoq3Parser.Val_strContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_str.
    def exitVal_str(self, ctx:eoq3Parser.Val_strContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_str_simple.
    def enterVal_str_simple(self, ctx:eoq3Parser.Val_str_simpleContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_str_simple.
    def exitVal_str_simple(self, ctx:eoq3Parser.Val_str_simpleContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_str_quote.
    def enterVal_str_quote(self, ctx:eoq3Parser.Val_str_quoteContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_str_quote.
    def exitVal_str_quote(self, ctx:eoq3Parser.Val_str_quoteContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_dat.
    def enterVal_dat(self, ctx:eoq3Parser.Val_datContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_dat.
    def exitVal_dat(self, ctx:eoq3Parser.Val_datContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_non.
    def enterVal_non(self, ctx:eoq3Parser.Val_nonContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_non.
    def exitVal_non(self, ctx:eoq3Parser.Val_nonContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_lst.
    def enterVal_lst(self, ctx:eoq3Parser.Val_lstContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_lst.
    def exitVal_lst(self, ctx:eoq3Parser.Val_lstContext):
        pass


    # Enter a parse tree produced by eoq3Parser#val_qry.
    def enterVal_qry(self, ctx:eoq3Parser.Val_qryContext):
        pass

    # Exit a parse tree produced by eoq3Parser#val_qry.
    def exitVal_qry(self, ctx:eoq3Parser.Val_qryContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_seg.
    def enterQry_seg(self, ctx:eoq3Parser.Qry_segContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_seg.
    def exitQry_seg(self, ctx:eoq3Parser.Qry_segContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_sym.
    def enterQry_sym(self, ctx:eoq3Parser.Qry_symContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_sym.
    def exitQry_sym(self, ctx:eoq3Parser.Qry_symContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_sym_short.
    def enterQry_sym_short(self, ctx:eoq3Parser.Qry_sym_shortContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_sym_short.
    def exitQry_sym_short(self, ctx:eoq3Parser.Qry_sym_shortContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_sym_his.
    def enterQry_sym_his(self, ctx:eoq3Parser.Qry_sym_hisContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_sym_his.
    def exitQry_sym_his(self, ctx:eoq3Parser.Qry_sym_hisContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_sym_other.
    def enterQry_sym_other(self, ctx:eoq3Parser.Qry_sym_otherContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_sym_other.
    def exitQry_sym_other(self, ctx:eoq3Parser.Qry_sym_otherContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_sym_full.
    def enterQry_sym_full(self, ctx:eoq3Parser.Qry_sym_fullContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_sym_full.
    def exitQry_sym_full(self, ctx:eoq3Parser.Qry_sym_fullContext):
        pass


    # Enter a parse tree produced by eoq3Parser#qry_arg.
    def enterQry_arg(self, ctx:eoq3Parser.Qry_argContext):
        pass

    # Exit a parse tree produced by eoq3Parser#qry_arg.
    def exitQry_arg(self, ctx:eoq3Parser.Qry_argContext):
        pass



del eoq3Parser


'''
 Bjoern Annighoefer 2024
'''

from .cmdstream import EoqFileInStream

def ValidateEoqFile(infile:str)->None:
    '''Validates the syntax of an EOQ file
    '''
    tester = EoqFileInStream()
    tester.Begin()
    tester.LoadEoqFile(infile)
    tester.Flush()


'''
 CmdStream are instances that produce, transform or consume a sequence of commands

 Bjoern Annighoefer 2024
'''

from ...domain import Domain
from ...value import VAL, U32, I64, STR, LST, NON
from ...query import Qry, Obj, His, IsObj, IsHis, HisId
from ...command import CMD_TYPES, Cmd, Crt, Get, Upd
from ...concepts import *
from ...serializer import CreateSerializer
from ...error import EOQ_ERROR_INVALID_VALUE

from .eoqfileformat import CMD_SEPERATOR, LINE_COMMENT_START
from .resnamegen import ResNameGenA

from typing import Tuple, List, Dict, Set, Hashable

class CmdStreamA:
    '''Abstract base class of all command streams.
    A command stream process a series of commands.
    Each command is a tuple of command and result.
    Not all results must be real results.
    
    Different streams specializations can have 
    different pre and postconditions on streams.
    The precondition is what properties a stream must
    exhibit to be processible by the stream instance.
    Postconditions describe the stream after the 
    stream instance.
    Stream properties are described by
    - ordered: is the sequence of commands ordered such 
    that all dependencies are resolved, i.e. are 
    all elements created before used. 
    - resName: are res names assigned to cmds 
    - res: are results assigned to cmds
    - refByHis: His is used in cmds
    - refByObj: Obj is used in cmds

    stream pre and postconditions must
    match when connecting streams.

    The typical usage of a stream is 
    Begin()
    ... do something with the stream ...
    Flush()
    
    Multiple streams can be connected by calling
    Connect(next stream instance).
    Connected stream instance invove each other 
    in the connect order with the commands passing 
    by. The order of commands and calling time 
    can however be varied by some streams, like 
    dependency sorting streams.
    
    '''
    def __init__(self):
        raise NotImplementedError()
    
    def Begin(self)->None:
        '''Used to init or re-init a cmd stream. 
        Should be called before using it.
        '''
        raise NotImplementedError()
    
    def Flush(self):
        '''Force the stream to release all pending commands
        '''
        raise NotImplementedError()
    
    def PushCmd(self, cmd:Cmd, res:VAL=NON())->None:
        '''Called by the user of the writer to add a new command to the stream
        '''
        raise NotImplementedError()
        
    def WriteComment(self, comment:str):
        '''Called by the user of the writer to insert a new comment
        '''
        raise NotImplementedError()
    
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        '''Is called for every resolved command.
        Should return the a the (modified) command and result if applicable otherwise the input elements.
        Override!
        '''
        raise NotImplementedError()
    
    def OnComment(self, comment:str)->str:
        '''Is called for every released comment.
        This shall return the the (modified) comment.
        Override!
        '''
        raise NotImplementedError()
    
    def Connect(self, streamToConnect)->None:
        '''Add a stream to be called with the commands from this stream
        '''
        raise NotImplementedError()
        
    #@Override
    def Disconnect(self, streamToDisconnect)->None:
        '''Remove a command stream from the stream chain
        '''
        raise NotImplementedError()

class CmdStream(CmdStreamA):
    ''' Basic command stream. 
    Implements the basic function of a command stream:
    forwards all given commands and comments to connected streams.
    Does not modify stream properties.
    Can be extended to create more sophisticated command streams.
    
    Stream preconditions:
    - ordered: any
    - resName: any
    - res: any
    - refByHis: any
    - refByObj: any
    
    Stream postconditions:
    - ordered: unchanged
    - resName: unchanged
    - res: unchanged
    - refByHis: unchanged
    - refByObj: unchanged
    '''
    def __init__(self):
        self.nextStreams:List[CmdStreamA] = []
    
    def Begin(self):
        ''' Calls begin for all connected streams
        '''
        for s in self.nextStreams:
            s.Begin()
    
    #@Override  
    def Flush(self):
        '''Flushes all connected command streams
        '''
        for s in self.nextStreams:
            s.Flush()
    
    #@Override
    def PushCmd(self, cmd:Cmd, res:VAL=NON()):
        '''Called by the user of the writer to indicate a new command to be written
        '''
        self._ReleaseCmd(cmd, res)
    
    #@Override  
    def PushComment(self, comment:str):
        '''Called by the user of the writer to insert a new comment
        '''
        self._ReleaseComment(comment)
    
    #@Override  
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        '''Just forwards command and result unmodified.
        Override to change behavior.
        '''
        return (cmd,res)
    
    #@Override  
    def OnComment(self, comment:str)->str:
        ''' Just returns the comment.
        Override to change behavior.
        '''
        return comment
    
    #@Override
    def Connect(self, streamToConnect:CmdStreamA)->None:
        '''Add a stream to be called with the commands from this stream
        '''
        if(streamToConnect not in self.nextStreams):
            self.nextStreams.append(streamToConnect)
        
    #@Override
    def Disconnect(self, streamToDisconnect:CmdStreamA)->None:
        '''Remove a command stream from the stream chain
        '''
        if(streamToDisconnect in self.nextStreams):
            self.nextStreams.remove(streamToDisconnect)
            
    ### INTERNALS ####
            
    def _ReleaseCmd(self, cmd:Cmd, res:VAL)->VAL:
        '''Forwards a cmd to OnCmd and all connected streams.
        Connected streams are called in the order they where connected.
        Should be called within the cmd stream to indicate a cmd 
        is ready for further processing.
        '''
        (cmd,res) = self.OnCmd(cmd, res)
        for s in self.nextStreams:
            s.PushCmd(cmd,res)
        return res
            
    def _ReleaseComment(self, comment:str)->None:
        '''Forwards a cmd to OnCmd and all connected streams.
        Connected streams are called in the order they where connected.
        Should be called within the cmd stream to indicate a cmd 
        is ready for further processing.
        '''
        comment = self.OnComment(comment)
        for s in self.nextStreams:
            s.PushComment(comment)
            
class RecorderCmdStream(CmdStream):
    '''Stores all pushed cmds into an internal list.
    The list can be retrieved by GetCmds().
    
    Stream preconditions:
    - ordered: any
    - resName: any
    - res: any
    - refByHis: any
    - refByObj: any
    
    Stream postconditions:
    - ordered: unchanged
    - resName: unchanged
    - res: unchanged
    - refByHis: unchanged
    - refByObj: unchanged
    '''
    def __init__(self):
        super().__init__()
        self.cmds:List[Cmd] = None
    
    #@Override    
    def Begin(self):
        '''Restarts recording
        '''
        self.cmds = []
        CmdStream.Begin(self)
    
    #@Override
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        ''' just stores commands in an internal list'''
        self.cmds.append(cmd)
        return (cmd,res)
    
    def GetCmds(self)->List[Cmd]:
        '''Returns a non writable copy of all commands
        streamed since last Begin() call.
        '''
        return list(self.cmds)
        
class DomainWrapperStream(CmdStream,Domain):
    '''Wraps a domain and forwards all commands to it, while also forwarding commands to connected streams.
    This stream is useful, for instance, to record all commands while also executing them on a domain.
    Event forwarding is not supported.

    Stream preconditions:
    - ordered: no
    - resName: no
    - res: no
    - refByHis: yes
    - refByObj: yes

    Stream postconditions:
    - ordered: unchanged
    - resName: unchanged
    - res: yes if execution successful
    - refByHis: unchanged
    - refByObj: unchanged

    '''
    def __init__(self, domain:Domain):
        CmdStream.__init__(self)
        Domain.__init__(self)
        self.domain = domain

    #@Override
    def Do(self, cmd:Cmd, sessionId:str=None, asDict:bool=False, readOnly=False)->VAL:
        '''Executes a command on the domain
        '''
        #first execute the command
        res = self.domain.Do(cmd,sessionId,asDict,readOnly)
        #second forward the command to connected streams
        self._ReleaseCmd(cmd,res)
        return res

    #@Override
    def PushCmd(self, cmd:Cmd, res:VAL=NON()):
        ''' Not supported
        '''
        raise NotImplementedError()
            
class DependencyResolverStream(CmdStream):
    '''A command stream implementation, sorting comments according to their dependencies and issue
    commands only for writing, if dependencies have been resolved.
    
    Stream preconditions:
    - ordered: no
    - resName: yes if resolveHis
    - res: yes if resolveObj
    - refByHis: yes if resolveHis
    - refByObj: yes if resolveObj
    
    Stream postconditions:
    - ordered: yes
    - resName: unchanged
    - res: unchanged
    - refByHis: unchanged
    - refByObj: unchanged
    
    '''
    def __init__(self, resolveObj:bool=True, resolveHis:bool=True):
        super().__init__()
        self.resolveObj = resolveObj
        self.resolveHis = resolveHis
        #internals
        self.cmdQueue:List[Tuple[Cmd,VAL,Set[Hashable]]] = None
        self.resolvedDependencies:Set[Hashable]= None #stores references already seen
        
    #@Override
    def Begin(self)->None:
        self.cmdQueue = []
        self.resolvedDependencies = set()
        super().Begin()
        
    #@Override
    def Flush(self)->None:
        nUnresolvedCmds = len(self.cmdQueue)
        if(0<nUnresolvedCmds):
            print("Warning: %d commands have unresolved dependencies."%(nUnresolvedCmds))
            self._ReleaseComment("Warning: %d commands have unresolved dependencies:"%(nUnresolvedCmds))
            for c in self.cmdQueue:
                self._ReleaseComment(str(c))
        super().Flush()
    
    #@Override
    def PushCmd(self, cmd:Cmd, res:VAL=NON())->None:
        '''target is the element this command creates
        '''
        unresolved = self.__CollectUnresolvedCmdDepencies(cmd)
        if(0<len(unresolved)):
            #cmd cannot be processed now. command contains unknown elements. Que it to resolve later
            self.cmdQueue.append((cmd,res,unresolved))
        else:
            rewRes = self.__WriteCmdAndUpdateLut(cmd, res)
            while(None != rewRes):
                rewRes = self.__ResolveAndWriteQueuedCmds(rewRes,cmd.r)
    
    ### INTERNALS ###
     
    def __WriteCmdAndUpdateLut(self, cmd:Cmd, res:VAL)->Obj:
        '''Returns if new a new element was added to the LUT
        '''
        if(CMD_TYPES.CRT == cmd.cmd):
            #in case this is an CRT, a new entry could be created in the LUT
            res = self._ReleaseCmd(cmd, res)
            self.__ResToLut(res, cmd.r)
            return res
        else:
            self._ReleaseCmd(cmd, res) 
            return None
        
    def __ResToLut(self, res:Obj, resName:str)->None:
        if(self.resolveObj):
            self.resolvedDependencies.add(res)
        if(self.resolveHis and None != resName):
            self.resolvedDependencies.add(His(resName))
    
    def __ResolveAndWriteQueuedCmds(self, newRes:Obj, newResName:str)->Obj:
        '''Returns true if another cmd causing a LUT was removed
        '''
        rewRes = None
        writtenCmds = []
        #process list until LUT needs to be updated
        for i in range(len(self.cmdQueue)):
            cmd, res, unresolved = self.cmdQueue[i]
            if(newRes in unresolved):
                unresolved.remove(newRes)
            if(None != newResName):
                if(His(newResName) in unresolved):
                    unresolved.remove(His(newResName))
        #after full loop is completed we can look for releasable cmds
        for i in range(len(self.cmdQueue)):
            cmd, res, unresolved = self.cmdQueue[i]
            if(0==len(unresolved)):
                rewRes = self.__WriteCmdAndUpdateLut(cmd,res)
                writtenCmds.append(i)
                if(None != rewRes):
                    break #if the LUT was updated we need to break and restart
        #remove all cmds from list that have been written
        for i in range(len(writtenCmds)):
            self.cmdQueue.pop(writtenCmds[i]-i) #-i corrects the index for already popped commands
        return rewRes
    
    def __CollectUnresolvedCmdDepencies(self, cmd:Cmd)->Set[Hashable]:
        '''See if all Obj in a command are known and replace them.
        Returns if all elements are replaced
        '''
        unresolved = self.__CollectUnresolvedListDepenciencies(cmd.a)
        return unresolved
    
    def __CollectUnresolvedListDepenciencies(self, l:Tuple[list,LST])->Set[Hashable]:
        unresolved = set()
        for i in range(len(l)):
            a = l[i]
            if(self.resolveObj and IsObj(a)):
                if(a not in self.resolvedDependencies):
                    unresolved.add(a)
            if(self.resolveHis and IsHis(a)):
                if(a not in self.resolvedDependencies):
                    unresolved.add(a)
            if(LST == type(a)):
                unresolved = unresolved.union(self.__CollectUnresolvedListDepenciencies(a))
            else:
                pass #all other values remain unconsidered
        return unresolved
    
    
class RefTypeToggleStream(CmdStream):
    ''' Utility class to change the reference type of an ordered 
    cmd stream from Obj to His or vise versa
    
    Stream preconditions:
    - ordered: yes
    - resName: yes if refByHis
    - res: yes if refByObj
    - refByHis: no if refByObj=yes
    - refByObj: no if refByHis=yes
    
    Stream postconditions:
    - ordered: unchanged
    - resName: unchanged
    - res: replaced by domain Obj
    - refByHis: replaced by domain Obj
    - refByObj: replaced by domain Obj
    
    '''
    def __init__(self, resolveObj:bool=False, resolveHis:bool=True):
        super().__init__()
        if(resolveObj and resolveHis):
            raise EOQ_ERROR_INVALID_VALUE("resolveObj and resolveHis cannot both be true.")
        self.resolveObj = resolveObj
        self.resolveHis = resolveHis
        #internals
        self.hisLut:Dict[str,Obj] = None
        self.resLut:Dict[Obj,Hashable] = None
        
    #Override
    def Begin(self):
        self.hisLut = {}
        self.resLut = {}
        super().Begin()
        
    #@Override   
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        ''' executes a given command on a domain.
        res is overwritten with the result from the domain.
        '''
        self._ResolveCmd(cmd)
        if(CMD_TYPES.CRT == cmd.cmd and None != cmd.r): #for CRT store the new results
            if(self.resolveObj):
                self.resLut[res] = His(cmd.r)
            elif(None != cmd.r):
                self.hisLut[cmd.r] = res
        return (cmd,res)
    
    ### INTERNALS ###
    
    def _ResolveCmd(self, cmd:Cmd)->None:
        '''See if all Obj in a command are known and replace them.
        Returns if all elements are replaced
        '''
        self.__ResolveList(cmd.a)
    
    def __ResolveList(self, l:Tuple[list,LST])->None:
        for i in range(len(l)):
            a = l[i]
            if(self.resolveObj and IsObj(a)):
                if(a in self.resLut):
                    l[i] = self.resLut[a]
                else:
                    raise EOQ_ERROR_INVALID_VALUE("Cannot resolve %s."%(a))
            elif(self.resolveHis and IsHis(a)):
                resName = HisId(a)
                if(resName in self.hisLut):
                    l[i] = self.hisLut[resName]
                else:
                    raise EOQ_ERROR_INVALID_VALUE("Cannot resolve %s."%(a))
            if(LST == type(a)):
                self.__ResolveList(a)
            else:
                pass #all other values remain unchanged
    
    
class ObjToHisStream(RefTypeToggleStream):
    '''Translates all Obj to His elements if res and resnames are given.
    Stream preconditions:
    - ordered: yes
    - resName: yes
    - res: yes
    - refByHis: no
    - refByObj: yes
    
    Stream postconditions:
    - ordered: yes
    - resName: yes
    - res: yes
    - refByHis: yes
    - refByObj: no
        
    '''
    def __init__(self):
        super().__init__(True,False)
        
class HisToObjStream(RefTypeToggleStream):
    '''Translates all Obj to His elements if res and resnames are given.
    Stream preconditions:
    - ordered: yes
    - resName: yes
    - res: yes
    - refByHis: yes
    - refByObj: no
    
    Stream postconditions:
    - ordered: yes
    - resName: yes
    - res: yes
    - refByHis: no
    - refByObj: yes
        
    '''
    def __init__(self):
        super().__init__(False,True)
        

class ResNameGenStream(CmdStream):
    '''Changes the res names in a cmd stream by applying a ResNameGen to all elements.
    
    Stream preconditions:
    - ordered: no
    - resName: no
    - res: partially as required by generator 
    - refByHis: any
    - refByObj: any
    
    Stream postconditions:
    - ordered: unchanged
    - resName: partially, depending on used generator 
    - res: unchanged
    - refByHis: unchanged
    - refByObj: unchanged
    
    '''
    def __init__(self, generator:ResNameGenA, inventForEmptyRes:bool = False):
        '''
        Empty res are skipped. 
        If inventForEmptyRes is true, than invent names for those as well.
        '''
        super().__init__()
        self.generator = generator
        self.inventForEmptyRes = inventForEmptyRes
        
    #@Override   
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        resName = cmd.r
        if(res.IsNone()):
            if(self.inventForEmptyRes):
                resName = self.generator.InventResName()
            else: 
                pass #do not generate a name, but keep the old value
        else:
            resName = self.generator.GenResName(res)
        cmd.r = resName
        return (cmd, res)

### DOMAIN STREAMS

class ToDomainCmdStream(RefTypeToggleStream):
    ''' Utility class to write a sequence of commands sequentially directly to a domain
    
    Stream preconditions:
    - ordered: yes
    - resName: yes if refByHis
    - res: yes if refByObj
    - refByHis: no if refByObj=yes
    - refByObj: no if refByHis=yes
    
    Stream postconditions:
    - ordered: unchanged
    - resName: unchanged
    - res: replaced by domain Obj
    - refByHis: replaced by domain Obj
    - refByObj: replaced by domain Obj
    
    '''
    def __init__(self, domain:Domain, sessionId:str, resolveObj:bool=False, resolveHis:bool=True):
        super().__init__(resolveObj,resolveHis)
        self.domain = domain
        self.sessionId = sessionId
        
    #@Override   
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        ''' executes a given command on a domain.
        res is overwritten with the result from the domain.
        '''
        if(self.resolveHis or self.resolveObj):
            self._ResolveCmd(cmd)
            newRes = self.domain.Do(cmd,self.sessionId)
            if(CMD_TYPES.CRT == cmd.cmd): #for CRT store the new results
                if(IsObj(res)):
                    self.resLut[res] = newRes
                if(None != cmd.r):
                    self.hisLut[cmd.r] = newRes
        else:
            newRes = self.domain.Do(cmd,self.sessionId)
        return (cmd,newRes)
    
    
class FromDomainCmdStream(CmdStream):
    '''Converts a model into commands
    '''
    def __init__(self, domain:Domain, sessionId:str):
        super().__init__()
        self.domain = domain
        self.sessionId = sessionId
            
    def LoadElement(self, root:Obj):
        concept = self.domain.Do( Get(Qry(root).Pth(STR(MXELEMENT.CONCEPT))) )
        if(CONCEPTS.M2MODEL==concept):
            self.LoadM2Model(root)
        elif(CONCEPTS.M1MODEL==concept):
            self.LoadM1Model(root)
        elif(CONCEPTS.MXMDB==concept):
            self.LoadMxMdb(root)
        else:
            raise EOQ_ERROR_INVALID_VALUE("root must be *M1MODEL, *M2MODEL or *MDB")
        
    def LoadMxMdb(self,  target:Obj)->None:
        m2models = self.domain.Do( Get( Qry(target).Pth(STR(MXMDB.M2MODELS))),self.sessionId)
        for m in m2models:
            self.LoadM2Model(m)
        m1models = self.domain.Do( Get( Qry(target).Pth(STR(MXMDB.M1MODELS))),self.sessionId)
        for m in m1models:
            self.LoadM1Model(m)
        
    def LoadM1Model(self, target:Obj)->None:
        name = self.domain.Do( Get( Qry(target).Pth(STR(M1MODEL.M2MODEL)).Pth(STR(MXELEMENT.STRID))),self.sessionId)
        self.__ConceptToFileCmdStream(target,CONCEPTS.M1MODEL,[M1MODEL.M2MODEL,M1MODEL.NAME],self.__MxElementToCmdStream, {M1MODEL.M2MODEL:name})
        objects = self.domain.Do( Get( Qry(target).Pth(STR(M1MODEL.OBJECTS))),self.sessionId)
        for o in objects:
            name = self.domain.Do( Get( Qry(o).Pth(STR(M1OBJECT.M2CLASS)).Pth(STR(MXELEMENT.STRID))),self.sessionId)
            self.__ConceptToFileCmdStream(o,CONCEPTS.M1OBJECT,[M1OBJECT.M2CLASS, M1OBJECT.MODEL, M1OBJECT.NAME],self.__MxElementToCmdStream, {M1OBJECT.M2CLASS:name})
            attributes = self.domain.Do( Get( Qry(o).Pth(STR(M1OBJECT.ATTRIBUTES))),self.sessionId)
            for f in attributes:
                name = self.domain.Do( Get( Qry(f).Pth(STR(M1ATTRIBUTE.M2ATTRIBUTE)).Pth(STR(M2ATTRIBUTE.NAME))),self.sessionId)
                self.__ConceptToFileCmdStream(f,CONCEPTS.M1ATTRIBUTE,[M1ATTRIBUTE.M2ATTRIBUTE,M1ATTRIBUTE.OBJECT,M1ATTRIBUTE.VALUE],self.__MxElementToCmdStream, {M1ATTRIBUTE.M2ATTRIBUTE:name})
            associations = self.domain.Do( Get( Qry(o).Pth(STR(M1OBJECT.SRCASSOCIATIONS))),self.sessionId)
            for f in associations:
                name = self.domain.Do( Get( Qry(f).Pth(STR(M1ASSOCIATION.M2ASSOCIATION)).Pth(STR(M2ASSOCIATION.DSTNAME))),self.sessionId)
                self.__ConceptToFileCmdStream(f,CONCEPTS.M1ASSOCIATION,[M1ASSOCIATION.M2ASSOCIATION,M1ASSOCIATION.SRC,M1ASSOCIATION.DST],self.__MxElementToCmdStream, {M1ASSOCIATION.M2ASSOCIATION:name})
            compositions = self.domain.Do( Get( Qry(o).Pth(STR(M1OBJECT.PARENTCOMPOSITIONS))),self.sessionId)
            for f in compositions:
                name = self.domain.Do( Get( Qry(f).Pth(STR(M1COMPOSITION.M2COMPOSITION)).Pth(STR(M2COMPOSITION.NAME))),self.sessionId)
                self.__ConceptToFileCmdStream(f,CONCEPTS.M1COMPOSITION,[M1COMPOSITION.M2COMPOSITION,M1COMPOSITION.PARENT,M1COMPOSITION.CHILD],self.__MxElementToCmdStream, {M1COMPOSITION.M2COMPOSITION:name})
    
    def LoadM2Model(self, target:Obj)->None:
        self.__ConceptToFileCmdStream(target,CONCEPTS.M2MODEL,[M2PACKAGE.NAME],self.__M2PackageFeaturesToCmdStream)
    
    def __MxElementToCmdStream(self, target:Obj)->None:
        docu = self.domain.Do( Get( Qry(target).Pth(STR(MXELEMENT.DOCUMENTATION))),self.sessionId)
        if(not docu.IsNone()):
            self.PushCmd(Upd(target,STR(MXELEMENT.DOCUMENTATION),docu), NON())
        owner = self.domain.Do( Get( Qry(target).Pth(STR(MXELEMENT.OWNER))),self.sessionId)
        if(not owner.IsNone()):
            self.PushCmd(Upd(target,STR(MXELEMENT.OWNER),owner), NON())
        group = self.domain.Do( Get( Qry(target).Pth(STR(MXELEMENT.GROUP))),self.sessionId)
        if(not owner.IsNone()):
            self.PushCmd(Upd(target,STR(MXELEMENT.GROUP),group), NON())
        permissions = self.domain.Do( Get( Qry(target).Pth(STR(MXELEMENT.PERMISSIONS))),self.sessionId)
        for p in permissions:
            self.PushCmd(Upd(target,STR(MXELEMENT.PERMISSIONS),p,I64(-1)), NON())
        constraints = self.domain.Do( Get( Qry(target).Pth(STR(MXELEMENT.CONSTRAINTS))),self.sessionId)
        for c in constraints:
            expression = self.domain.Do(Get( Qry(c).Pth(STR(MXCONSTRAINT.EXPRESSION))),self.sessionId)
            self.PushCmd(Crt(STR(CONCEPTS.MXCONSTRAINT),U32(1),LST([target,expression])), NON())
            self.__MxElementToCmdStream(c)
        
    def _M2PackageToFile(self, target:Obj)->None:
        self.__ConceptToFileCmdStream(target,CONCEPTS.M2PACKAGE,[M2PACKAGE.NAME,M2PACKAGE.SUPERPACKAGE],self.__M2PackageFeaturesToCmdStream)
        
    def __M2PackageFeaturesToCmdStream(self, target:Obj)->None:
        self.__MxElementToCmdStream(target)
        enums = self.domain.Do( Get( Qry(target).Pth(STR(M2PACKAGE.ENUMS))),self.sessionId)
        for e in enums:
            self.__ConceptToFileCmdStream(e,CONCEPTS.M2ENUM,[M2ENUM.NAME,M2ENUM.PACKAGE],self.__MxElementToCmdStream)
            options = self.domain.Do( Get( Qry(e).Pth(STR(M2ENUM.OPTIONS))),self.sessionId)
            for o in options:
                self.__ConceptToFileCmdStream(o,CONCEPTS.M2OPTIONOFENUM,[M2OPTIONOFENUM.NAME,M2OPTIONOFENUM.VALUE,M2OPTIONOFENUM.ENUM],self.__MxElementToCmdStream)
        classes = self.domain.Do( Get( Qry(target).Pth(STR(M2PACKAGE.CLASSES))),self.sessionId)
        for c in classes:
            self.__ConceptToFileCmdStream(c,CONCEPTS.M2CLASS,[M2CLASS.NAME,M2CLASS.ISABSTRACT,M2CLASS.PACKAGE],self.__MxElementToCmdStream)
            inheritances = self.domain.Do( Get( Qry(c).Pth(STR(M2CLASS.MYGENERALIZATIONS))),self.sessionId)
            for i in inheritances:
                self.__ConceptToFileCmdStream(i,CONCEPTS.M2INHERITANCE,[M2INHERITANCE.SUBCLASS,M2INHERITANCE.SUPERCLASS],self.__MxElementToCmdStream)
            attributes = self.domain.Do( Get( Qry(c).Pth(STR(M2CLASS.MYATTRIBUTES))),self.sessionId)
            for f in attributes:
                self.__ConceptToFileCmdStream(f,CONCEPTS.M2ATTRIBUTE,[M2ATTRIBUTE.NAME,M2ATTRIBUTE.CLASS,M2ATTRIBUTE.PRIMTYPE,M2ATTRIBUTE.MUL,M2ATTRIBUTE.UNIT,M2ATTRIBUTE.ENUM],self.__MxElementToCmdStream)
            associations = self.domain.Do( Get( Qry(c).Pth(STR(M2CLASS.MYSRCASSOCIATIONS))),self.sessionId)
            for f in associations:
                self.__ConceptToFileCmdStream(f,CONCEPTS.M2ASSOCIATION,[M2ASSOCIATION.SRCNAME,M2ASSOCIATION.SRCCLASS,M2ASSOCIATION.SRCMUL,M2ASSOCIATION.DSTNAME,M2ASSOCIATION.DSTCLASS,M2ASSOCIATION.DSTMUL,M2ASSOCIATION.ANYDST],self.__MxElementToCmdStream)
            compositions = self.domain.Do( Get( Qry(c).Pth(STR(M2CLASS.MYPARENTCOMPOSITIONS))),self.sessionId)
            for f in compositions:
                self.__ConceptToFileCmdStream(f,CONCEPTS.M2COMPOSITION,[M2COMPOSITION.NAME,M2COMPOSITION.PARENTCLASS,M2COMPOSITION.CHILDCLASS,M2COMPOSITION.MULCHILD,M2COMPOSITION.ANYCHILD],self.__MxElementToCmdStream)
            packages = self.domain.Do( Get( Qry(target).Pth(STR(M2PACKAGE.SUBPACKAGES))),self.sessionId)
        for p in packages:
            self._M2PackageToFile(p)
    
    def __ConceptToFileCmdStream(self, target:Obj, conceptId:str, createArgNames:List[str], superClassCall=None, caReplacements:Dict[str,VAL]={}):
        createArgs = []
        for n in createArgNames:
            if(n in caReplacements):
                createArgs.append(caReplacements[n])
            else:
                createArgs.append(self.domain.Do( Get( Qry(target).Pth(STR(n))),self.sessionId))
        cmd = Crt(STR(conceptId),U32(1),LST(createArgs))
        self.PushCmd(cmd, target)
        if(None!=superClassCall):
            superClassCall(target)


### EOQFILE STREAMS ###
    
    
class EoqFileOutStream(CmdStream):
    ''' Utility class to write a sequence of commands sequentially to a file.
    
    Stream preconditions:
    - ordered: yes (only EOQ file shall contain ordered cmds)
    - resName: yes
    - res: any (is not used)
    - refByHis: yes (EOQ files are reference by His)
    - refByObj: no
    
    Stream postconditions:
    - ordered: unchanged
    - resName: unchanged
    - res: unchanged
    - refByHis: unchanged
    - refByObj: unchanged
    
    '''
    def __init__(self, outfile:str, serializerType:str = "TXT"):
        super().__init__()
        self.outfile = outfile
        self.serializer = CreateSerializer(serializerType)
    
    #@Override
    def Begin(self):
        self.file = open(self.outfile,"w")
        super().Begin()
        
    #@Override   
    def OnCmd(self, cmd:Cmd, res:VAL)->Tuple[Cmd,VAL]:
        strCmd = self.serializer.SerCmd(cmd)
        self.file.write(strCmd)
        self.file.write(CMD_SEPERATOR)
        return (cmd, res)
    
    #@Override 
    def OnComment(self, comment:str)->str:
        self.file.write(LINE_COMMENT_START)
        self.file.write(comment)
        self.file.write(CMD_SEPERATOR)
        return comment
    
    #@Override  
    def Flush(self)->None:
        if(self.file):
            self.file.close()
        super().Flush()
        
        
class EoqFileInStream(CmdStream):
    ''' Utility class to read a stream of commands sequentially from a file.
    
    Stream preconditions:
    - (shall not be used after other stream)
    
    Stream postconditions:
    - ordered: yes (only EOQ file shall contain ordered cmds)
    - resName: yes
    - res: any (is not used)
    - refByHis: yes (EOQ files are reference by His)
    - refByObj: no
    
    '''
    def __init__(self, serializerType:str="TXT"):
        super().__init__()
        self.serializer = CreateSerializer(serializerType)
        
    def LoadEoqFile(self, infile:str)->None:
        f = open(infile,"r")
        cmdStr = '' #stores command string until it is complete
        lineNb = 0;
        try:
            for line in f:
                lineNb += 1
                if(line.startswith(LINE_COMMENT_START)): #comment
                    self.PushComment(line[len(LINE_COMMENT_START):])
                else:
                    if(line.endswith(CMD_SEPERATOR)):
                        cmdStr += line
                        #command is complete, so execute
                        cmdStr = cmdStr[:-2]
                        cmd = self.serializer.DesCmd(cmdStr)
                        self.PushCmd(cmd)
                        #empty buffer
                        cmdStr = ''
                    else:
                        cmdStr += line
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Eoq file load failed in %s line %d: %s"%(infile, lineNb, str(e)))
        finally:
            f.close()



from .serialdomain import SerialDomain
from .domaintoprocesswrapper import DomainToProcessWrapper
from .multiprocessingdomainclient import MultiprocessingQueueDomainClient
from .multiprocessingdomainhost import MultiprocessingQueueDomainHost
from .domainpool import DomainPool
from .remotedomainclient import RemoteDomainClient



'''
2019 Bjoern Annighoefer
'''
from .logger import DEFAULT_LOGGER_LEVELS,RegisterLogger
from .compoundlogger import CompoundLogger
from .consolelogger import ConsoleLogger
from .filelogger import FileLogger

import os
import logging

class ConsoleAndFileLogger(CompoundLogger):
    """A logger that outputs every thing to the console and dedicated files for each active log level
    """
    def __init__(self,activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING,logDir='./log',prefix=''):
        super().__init__(activeLevels)
        self.logDir = logDir
        self.AddLogger(ConsoleLogger(activeLevels))
        self.AddLogger(FileLogger(activeLevels,logDir,prefix))

RegisterLogger("CFL",ConsoleAndFileLogger,[ "logDir","prefix"])


'''
 2019 Bjoern Annighoefer
'''

from ..value import VAL,LST
from ..error import EoqErrorFactory,EOQ_ERROR_RUNTIME
from ..util.observable import Observable
from ..command import CMD_TYPES, Cmd, Res, Err
#type checking
from typing import Union

class Domain(Observable):
    def __init__(self):
        super().__init__()
    
    def Do(self, cmd:Cmd, sessionId:str=None, asDict:bool=False, readOnly=False)->VAL:
        '''
        Args:
        ----
        '''
        res = self.RawDo(cmd,sessionId,readOnly)
        if(CMD_TYPES.RES == res.cmd):
            val =  res.a[2]
            if(asDict): #return as dictionary works only for compound values
                if(0<len(res.a[3])):
                    valDct = {}
                    valNames = res.a[3]
                    valNameIndicies = res.a[4]
                    for i in range(len(valNames)):
                        valDct[valNames[i].GetVal()] = val[valNameIndicies[i]]
                    return valDct
                else:
                    raise EOQ_ERROR_RUNTIME("There are no value names in result. Cannot use asDict.")
            else: #return as list
                return val #the third entry is the value
        elif(CMD_TYPES.ERR == res.cmd):
            error = EoqErrorFactory(res.a[0].GetVal(), res.a[1].GetVal(), res.a[2].GetVal())
            raise error
        else:
            raise EOQ_ERROR_RUNTIME("Unexpected result type: %s"%(res.cmd))
        
    def RawDo(self, cmd:Cmd, sessionId:str=None, readOnly:bool=False)->Union[Res,Err]:
        '''Do function without converting the results back into the value
        '''
        raise NotImplemented()
        
    def DoAsync(self, cmd:Cmd, callbackSuccess:callable, callbackError:callable, sessionId:str=None, asDict:bool=False):
        raise NotImplemented()
            
    
    def Close(self):
        pass #is to be called if the domain closes down
    
                    
    
        
        


'''
 Bjoern Annighoefer 2019
'''

from .serializer import Serializer, RegisterSerializer

from ..value import VAL, BOL, U32, U64, I32, I64, F32, F64, STR, DAT, NON, QRY, LST
#import all queries because the eval function needs them
from ..query import Seg, Qry, Obj, Oni, His, Pth, Cls, Ino, Try, Met, Qrf, Stf, Slf, Not, Idx, Sel, Arr, Zip, Any, All, Equ, Eqa, Neq, Les, Gre, Rgx
#import all commands because the eval function needs them
from ..command import Cmd, Get, Upd, Del, Set, Add, Rem, Mov, Clo, Crt, Cpr, Mrg, Hel, Ses, Gby, Sts, Chg, Obs, Ubs, Msg, Ver, Gac, Res, Err, Evt, Scc, Cus, Cmp, GetCmdInfo
from ..frame import Frm
from ..error import EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_INVALID_VALUE


JS_FRM_PREFIX = "new FRM.";
JS_CMD_PREFIX = "new CMD.";
JS_QRY_PREFIX = "new QRY.";
JS_VAL_PREFIX = "new VAL.";

class JsSerializer(Serializer):
    '''Javascript Serializer   
    '''
    # @Overrides 
    def Name(self)->str:           
        return "JSC"
    # @Overrides 
    def SerFrm(self, frm:Frm)->str:
        return JS_FRM_PREFIX+"Frm('%s',%d,'%s','%s','%s',%s,'%s')"%(frm.typ,frm.uid,frm.ser,frm.dat.replace("\\","\\\\").replace("'","\\'"),frm.sid,"true" if frm.roc else "false",frm.ver);
    
    # @Override
    def SerCmd(self, cmd:Cmd)->str:
        return JS_CMD_PREFIX+self.__SerCmdRaw(cmd,0)
        
    def __SerCmdRaw(self, cmd:Cmd, level:int):
        if(isinstance(cmd,Cmp)):
            cmdStrs = ['.' + self.__SerCmdRaw(c,level+1) for c in cmd.a]
            if(0==level):
                return "Cmp("+self.__CmdMuteResNameStr(cmd,"[],")+")"+"".join(cmdStrs)
            else:
                return "Append("+JS_CMD_PREFIX+"Cmp("+self.__CmdMuteResNameStr(cmd,"[],")+")"+"".join(cmdStrs)+")"
        else:
            argStrs = [self.SerVal(a) for a in cmd.a]
            cmdInfo = GetCmdInfo(cmd.cmd)
            if(None != cmdInfo):
                if(cmdInfo.unpackArgs):
                    return "%s(%s%s)"%(self.__ToJsName(cmd.cmd),",".join(argStrs),self.__CmdMuteResNameStr(cmd,("," if 0<len(argStrs) else "")))
                else:
                    return "%s([%s]%s)"%(self.__ToJsName(cmd.cmd),",".join(argStrs),self.__CmdMuteResNameStr(cmd,","))
            else: #must be a custom command 
                return "Cus('%s',[%s]%s)"%(cmd.cmd,",".join(argStrs),self.__CmdMuteResNameStr(cmd,","))
        
    def __CmdMuteResNameStr(self, cmd:Cmd, prefix:str):
        segments = []
        if(cmd.m):
            segments.append("true")
        elif(cmd.r): #not mute but resname
            segments.append("false")
        if(cmd.r): #check a second time without the precondition that mute must be false
            segments.append("'%s'"%(cmd.r))
        if(0 < len(segments)):
            return prefix+",".join(segments)
        else:
            return ""
        
    # @Override    
    def SerQry(self,qry:Qry):
        segStrs = [self.SerSeg(s) for s in qry.v]
        return JS_QRY_PREFIX + '.'.join(segStrs)
    
    # @Override
    def SerSeg(self, seg:Seg):
        argStrs = [self.SerVal(a) for a in seg.v]
        return self.__ToJsName(seg.seg)+ '(' + ','.join(argStrs) + ')'
    
    # @Override    
    def SerVal(self, val:VAL):
        if(isinstance(val,NON)): 
            return JS_VAL_PREFIX + 'NON()'
        elif(isinstance(val,BOL)): 
            return JS_VAL_PREFIX + 'BOL(true)' if val.v else 'BOL(false)'
        elif(isinstance(val,U32)): 
            return JS_VAL_PREFIX + 'U32(%d)'%(val.v)
        elif(isinstance(val,U64)): 
            return JS_VAL_PREFIX + 'U64(%d)'%(val.v)
        elif(isinstance(val,I32)): 
            return JS_VAL_PREFIX + 'I32(%d)'%(val.v)
        elif(isinstance(val,I64)): 
            return JS_VAL_PREFIX + 'I64(%d)'%(val.v)
        elif(isinstance(val,F32)): 
            fstr = ('%f'%(val.v)).rstrip('0').rstrip('.') #remove trailing 0 and .
            return JS_VAL_PREFIX + 'F32(%s)'%(fstr) #remove trailing 0 and .
        elif(isinstance(val,F64)): 
            fstr = ('%f'%(val.v)).rstrip('0').rstrip('.') #remove trailing 0 and .
            return JS_VAL_PREFIX + 'F64(%s)'%(fstr) #remove trailing 0 and .
        elif(isinstance(val,STR)): 
            return JS_VAL_PREFIX + "STR('" + val.v.replace("\\","\\\\").replace("'","\\'") + "')"
        elif(isinstance(val,DAT)):
            us = ('%f'%(val.microsec())).rstrip('0').rstrip('.') #remove trailing 0 and .
            of = ('%f'%(val.utcoff())).rstrip('0').rstrip('.') #remove trailing 0 and .
            return JS_VAL_PREFIX + "DAT(%d,%d,%d,%d,%d,%d,%s,%s)"%(val.year(),val.month(),val.day(),val.hour(),val.minute(),val.second(),us,of)
        elif(isinstance(val,QRY)):
            return self.SerQry(val)
        elif(isinstance(val,LST)): #list must be the after str, because str is also a list
            valStrs = [self.SerVal(v) for v in val.v] #cannot use Val() here, because that decomposes the full list
            return JS_VAL_PREFIX + 'LST(['+','.join(valStrs)+'])'
        else:
            raise EOQ_ERROR_INVALID_TYPE("Unknown value type %s."%(type(val).__name__))
        
    def __ToJsName(self,typeStr:str)->str:
        return typeStr[0] + typeStr[1:].lower()

    def DesFrm(self, data)->Frm:
        try:
            pyexpr = self.__JsToPyConverter(data) #transform js to python
            return eval(pyexpr)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Frm: %s"%(e))
    
    def DesCmd(self, data)->Cmd:
        try:
            pyexpr = self.__JsToPyConverter(data) #transform js to python
            return eval(pyexpr)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Cmd: %s"%(e))
    
    def DesQry(self, data)->Qry:
        try:
            pyexpr = self.__JsToPyConverter(data) #transform js to python
            return eval(pyexpr)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Qry: %s"%(e))
    
    def DesSeg(self, data)->Seg:
        try:
            pyexpr = self.__JsToPyConverter(data) #transform js to python
            return eval(pyexpr)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Seg: %s"%(e))
    
    def DesVal(self, data):
        try:
            pyexpr = self.__JsToPyConverter(data) #transform js to python
            return eval(pyexpr)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Invalid Val: %s"%(e))
        
    def __JsToPyConverter(self, jsexpr:str)->str:
        ''' A very simple and non-safe converter of EOQ Javascript expressions to Python expressions
        '''
        jssegs = jsexpr.split("'")
        pysegs = []
        
        isQuotedStr = False
        for i in range(len(jssegs)):
            jsseg = jssegs[i]
            #find out if this is a quoted segment
            if(0==i):
                pass #the first segment is never quoted
            else:
                #count the number of \ prior to the '
                nBackslash = 0
                for j in range(len(pysegs[i-1])-1,-1,-1):
                    if(pysegs[i-1][j] == "\\"):
                        nBackslash += 1
                    else:
                        break;
                #only even backslashes terminate the quote
                if(isQuotedStr and nBackslash%2 == 1):
                    pass #no toggle, because a quoted str was hit
                else:
                    isQuotedStr = not isQuotedStr #normally, this toggles at every segment
            
            #depending on the type of segments replace keywords or not
            if(isQuotedStr):
                pysegs.append(jsseg)
            else:
                pysegs.append(jsseg.replace("true","True").replace("false","False").replace(JS_CMD_PREFIX,"").replace(JS_QRY_PREFIX,"").replace(JS_VAL_PREFIX,"").replace(JS_FRM_PREFIX,""))
        return "'".join(pysegs)
        
### make this serializer known globally 
RegisterSerializer(JsSerializer) #do not do this because it is a security risk


'''
The interface of the class the does the user verification and access control on element level. 
Different implementations are possible as long as they comply to this interface
This is also the base implementation that does pass all verifications

2022 Bjoern Annighoefer

'''
from ..value import STR, I64, PRM, VAL, NON, LST
from ..query import Obj
from ..mdb import ReadOnlyMdb

from typing import Union

class Verifier:
    def __init__(self):
        self.mdb:ReadOnlyMdb = None
    
    ### INIT AND CLOSE ###
    
    def Init(self, mdb:ReadOnlyMdb)->None:
        '''Initialize the Access controller and tell him the domain he is responsible for
        '''
        self.mdb = mdb
        
    def Deinit(self):
        '''Clean up the access controller
        '''
        pass
    
    ### PRE VERIFICATION ###
    # this functions are called before the data is changed
    
    def CreatePreVerify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        '''Raises an exception if the creation should be stopped
        '''
        pass
    
    def ReadPreVerify(self, target:Obj, featureName:STR, context:Obj=NON(), user:str=None)->None:
        '''Raises an exception if the intended read access should be stopped
        '''
        pass
    
    def UpdatePreVerify(self, target:Obj, featureName:STR, value:PRM, position:I64=I64(0), user:str=None)->None:
        '''Raises an exception if the intended update should be stopped
        '''
        pass
    
    
    def DeletePreVerify(self, target:Obj, user:str=None)->None:
        '''Raises an exception if the intended delete should be stopped
        '''
        pass
    
    ### POST VERIFICATION ###
    # this functions are called after the has changed.
    # if a failure is found here, a rollback of the previous change is necessary
    
    def CreatePostVerify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        '''Raises an exception if the creation needs to be undone.
        '''
        pass
    
#     def ReadPostVerify(self, target:Obj, featureName:STR, context:Obj)->None:
#         '''Raises an exception if the read needs to be undone.
#         '''
#         pass
    
    def UpdatePostVerify(self, target:Obj, featureName:STR, value:PRM, position:I64=I64(0), oldValue:PRM=NON(), user:str=None)->None:
        '''Raises an exception if the update needs to be undone.
        '''
        pass
    
    def DeletePostVerify(self, target:Obj, classId:STR, createArgs:LST, recoveryArgs:LST, user:str=None)->None:
        '''Raises an exception if the deletion needs to be undone.
        '''
        pass
    
    
    ### NOTIFIERS ####
    # Functions that inform the verifier that a certain action took place. 
    # This must not fail or raise exceptions, since all actions have passed pre and post verification beforehand
    
    def CreateNotify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        '''This is a callback informing the access controller that a create took place
        '''
        pass
    
    def ReadNotify(self, target:Obj, featureName:STR, context:Obj, val:VAL, user:str=None)->None:
        '''This is a callback informing the access controller that a read took place
        '''
        pass
    
    def UpdateNotify(self, target:Obj, featureName:STR, value:PRM, position:I64, oldValue:PRM, user:str=None)->None:
        '''This is a callback informing the access controller that an update took place
        '''
        pass
    
    def DeleteNotify(self, target:Obj, classId:STR, createArgs:LST, recoveryArgs:LST, user:str=None)->None:
        '''This is a callback informing the access controller that a deletion took place
        '''
        pass




from typing import Dict, List, Type

# BASIC ERROR CLASS
class EOQ_ERROR(Exception):
    ''' The class a exceptions thrown in the EOQ eco system inherit from
    
    '''
    def __init__(self, prefix:str, code:int, msg:str, trace:str=""):
        self.prefix = prefix
        self.code = code
        self.msg = msg
        self.trace = trace
        
    def __str__(self):
        return "%s(%d): %s %s"%(self.prefix,self.code,self.msg,"trace=%s"%(self.trace) if self.trace else "")

# EOQ ERROR REGISTRY
class EoqErrorInfo:
    ''' A class to store the error information in the EOQ error registry
    
    '''
    def __init__(self, code:int, description:str, errorClass):
        self.code = code
        self.description = description
        self.errorClass = errorClass
        
EOQ_ERROR_TYPE_REGISTRY:Dict[int,EoqErrorInfo] = {} # code -> EoqErrorInfo

def RegisterEoqErrorType(code:int, description:str, clazz:Type[EOQ_ERROR]):
    ''' Each new instance of EOQ should be registered in the error registry. 
    
    This enables automatic instantiation of the errors based on the code using the EoqErrorFactory.
    Each inherit error class must exhibit the same interface.
    
    '''
    if(code in EOQ_ERROR_TYPE_REGISTRY):
        raise EOQ_ERROR_RUNTIME('Error code %d is already registered.'%(code))
    else:
        EOQ_ERROR_TYPE_REGISTRY[code] = EoqErrorInfo(code, description, clazz)     




class EOQ_ERROR_CODES:
    ''' This is a general set of errors in EOQ that should be used for any error returned by eoq
    '''
    INVALID_TYPE           = 1 #not the the expected data type
    INVALID_VALUE          = 2 #not the expected value (e.g. range)
    INVALID_OPERATION      = 3 #operation is (currently) not possible
    DOES_NOT_EXIST         = 4 #no such element
    RUNTIME                = 5 #error during operation
    ACCESS_DENIED          = 6 #any failure resulting from access rights
    UNSUPPORTED            = 7 #something is not implemented or not supported with the current library
    UNKNOWN                = 9 #Something that should not happen and is not understood by EOQ
    
    
### SPECIALIZED ERROR CLASSES 
class EOQ_ERROR_INVALID_TYPE(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_INVALID_TYPE',EOQ_ERROR_CODES.INVALID_TYPE, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.INVALID_TYPE,'The data type used is not supported here.',EOQ_ERROR_INVALID_TYPE)

class EOQ_ERROR_INVALID_VALUE(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_INVALID_VALUE',EOQ_ERROR_CODES.INVALID_VALUE, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.INVALID_VALUE,'The value given is not expected.',EOQ_ERROR_INVALID_VALUE)

class EOQ_ERROR_INVALID_OPERATION(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_INVALID_OPERATION',EOQ_ERROR_CODES.INVALID_OPERATION, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.INVALID_OPERATION,'This operation is (currently) not possible.',EOQ_ERROR_INVALID_OPERATION)

class EOQ_ERROR_DOES_NOT_EXIST(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_DOES_NOT_EXIST',EOQ_ERROR_CODES.DOES_NOT_EXIST, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.DOES_NOT_EXIST,'Tried to access an non exiting element, attribute or index.',EOQ_ERROR_DOES_NOT_EXIST)

class EOQ_ERROR_RUNTIME(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_RUNTIME',EOQ_ERROR_CODES.RUNTIME, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.RUNTIME,'The operation failed.',EOQ_ERROR_RUNTIME)

class EOQ_ERROR_ACCESS_DENIED(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_ACCESS_DENIED',EOQ_ERROR_CODES.ACCESS_DENIED, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.ACCESS_DENIED,'Insufficient access rights.',EOQ_ERROR_ACCESS_DENIED)

class EOQ_ERROR_UNSUPPORTED(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_UNSUPPORTED',EOQ_ERROR_CODES.UNSUPPORTED, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.UNSUPPORTED,'The operation is not implemented.',EOQ_ERROR_UNSUPPORTED)

class EOQ_ERROR_UNKNOWN(EOQ_ERROR):
    def __init__(self, msg:str, trace:str=""):
        super().__init__('EOQ_ERROR_UNKNOWN',EOQ_ERROR_CODES.UNKNOWN, msg, trace)
RegisterEoqErrorType(EOQ_ERROR_CODES.UNKNOWN,'The reason for the error is not known.',EOQ_ERROR_UNKNOWN)



### ERROR FACTORY ###
        
def EoqErrorFactory(code:int, msg:str, trace:str='')->EOQ_ERROR:
    if(code in EOQ_ERROR_TYPE_REGISTRY):
        errorClass = EOQ_ERROR_TYPE_REGISTRY[code].errorClass
        return errorClass(msg, trace)
    else: 
        return EOQ_ERROR_UNKNOWN(code,"Unknown error type %d: %s"%(code, msg))
    
def EoqErrorAsString(error:EOQ_ERROR):
    return str(error)
    
        

    


'''
2019 Bjoern Annighoefer
'''

from .logger import Logger,RegisterLogger, DEFAULT_LOGGER_LEVELS, LOG_LEVELS_NAME_LUT

import os
import logging

class FileLogger(Logger):
    """A logger that logs to a file.
    Each log level has its own file.
    The log files are created in the logDir with the prefix and the log level as name.
    """
    def __init__(self,activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING,logDir:str='./log',prefix:str=''):
        super().__init__(activeLevels)
        self.logDir = logDir
        self.prefix = prefix #the text added infront of the log file name
        self.pyLoggers = {}
        #make sure the dir exists
        if(not os.path.isdir(self.logDir)):
            os.makedirs(self.logDir)
        #create native python loggers for each level
        for k,v in LOG_LEVELS_NAME_LUT.items():
            if(k & activeLevels):
                #init error python logger
                logger = logging.getLogger(v)
                logFile = os.path.join(self.logDir,"%s%s.log"%(self.prefix,v))
                fh = logging.FileHandler(logFile,'w')
                fh.setLevel(logging.INFO)
                formatter = logging.Formatter('%(asctime)s - %(message)s')
                fh.setFormatter(formatter)
                logger.addHandler(fh)
                logger.setLevel(logging.INFO)
                self.pyLoggers[v] = logger
                
    #@Override         
    def _Log(self,levelName:str,msg:str):
        self.pyLoggers[levelName].info(msg)

RegisterLogger("FIL",FileLogger,[ "logDir","prefix"])

